!
! Module: lccorr_lite_mod
!
! Authors:  Ian Stewart - SSC
!          
! Description: All this does is calculate the exposure correction for GTI edges falling within time bins. The task assumes that the source lies solely on ccd 1 (note this is NOT the central ccd for pn). There is no output as yet.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
!

module lccorr_lite_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use param  
  use errorHandling

  use fits_utils
  use dss

  implicit none 
  
  private 

  ! Command-line parameters:
  character(len=255) ::&
    rateSetName

  type :: localGtiType
    real(double) :: startTime, stopTime
  end type localGtiType

  real(double) :: timeBinWidth, startTimeFirstBin, stopTimeLastBin

  public lccorr_lite

contains 
  
  !---------------------------------------------------------------------
  subroutine lccorr_lite
  !---------------------------------------------------------------------

    implicit none

    real(double), pointer :: times(:)
    type(localGtiType), pointer :: gti(:)
    real(single), allocatable :: exposureFrac(:)
    type(DataSetT) :: set
    character(20) :: tableName = 'RATE'
    type(TableT) :: tab

    !--------------------------------------------------------------------
    ! Start:

    call readParameters

call dumpDSS(rateSetName)
return

    set = dataSet(rateSetName, READ)
    tab = table(set, tableName)
    call loadTimeSeries(tab, times)
    call loadGtis(set, gti)
    allocate(exposureFrac(size(times)))
    call calcExposure(times, gti, exposureFrac)
    deallocate(times, gti)
    call release(set)

  end subroutine lccorr_lite

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    rateSetName = stringParameter('srctsset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine loadTimeSeries(tab, times)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    real(double), pointer :: times(:)

    type(ColumnT) :: timeCol

    !-------------------------------------------------------------------
    ! Start

    timeBinWidth = real64Attribute(tab, 'TIMEDEL')
    timeCol = column(tab, 'TIME')
    times => real64Data(timeCol)

    startTimeFirstBin = times(0) - 0.5 * timeBinWidth
    stopTimeLastBin = times(size(times) - 1) + 0.5 * timeBinWidth

  end subroutine loadTimeSeries

  !---------------------------------------------------------------------
  subroutine loadGtis(set, gti)
  !---------------------------------------------------------------------
!*** now in my_eexpmap/dssaux_mod (shortly to be in ssclib).

!    use fitsutils_mod
!    use dss

    implicit none

    type(DataSetT), intent(in)  :: set
    type(localGtiType), pointer :: gti(:)

    type(BlockT)    :: inBlock
    type(DssT)      :: dssPointer
    logical         :: ccd1GtisFound
    integer         :: compNum, filtNum, numGtis
    type(DSfilterT) :: filter, ccd1GtiFilter
    character(255)  :: gtiBlockName
    type(TableT)    :: tab

    !-------------------------------------------------------------------
    ! Start

    inBlock = block(set, 0, READ)
    dssPointer = dataSubspace(inBlock)

    ! Find the dss component with CCDNR==1:
    ccd1GtisFound = .false.
    COMP: do compNum = 0, dssComponents(dssPointer) - 1
      ! See if there is a CCDNR filter:
      do filtNum = 0, dssFilters(dssComponent(dssPointer, compNum)) - 1
        filter = dssFilter(dssComponent(dssPointer, compNum), filtNum)
        if (dssFilterName(filter) .eq. 'CCDNR') then
    if (.not. valueIsPassedByDssFilter(1.0, filter)) cycle COMP
        end if
      end do

      ! See if there is also a GTI filter in this dss component:
      do filtNum = 0, dssFilters(dssComponent(dssPointer, compNum)) - 1
        if (dssFilterType(filter) .eq. GTIFILTERT) then
          ccd1GtisFound = .true.
          ccd1GtiFilter = dssFilter(dssComponent(dssPointer, compNum), filtNum)
        end if
      end do
    exit COMP
    end do COMP

    if (.not. ccd1GtisFound) then ! set 1 gti to cover the whole time series:
      allocate(gti(1))
      gti(1)%startTime = startTimeFirstBin - 1.0
      gti(1)%stopTime  = stopTimeLastBin   + 1.0
  return
    end if

    ! Get GTIs from the relevant dss component:
    gtiBlockName = dssFilterBlock(ccd1GtiFilter)
    tab = table(set, gtiBlockName)
    numGtis = numberOfRows(tab)
    allocate(gti(numGtis))
    call readColData(tab, 'START', gti%startTime)
    call readColData(tab, 'STOP',  gti%stopTime)

    call release(tab)
    call release(inBlock)

!*** check gtis are sorted, and gti(:)%stopTime always > gti(:)%startTime?

  end subroutine loadGtis

  !---------------------------------------------------------------------
  logical function valueIsPassedByDssFilter(value, filter)
  !---------------------------------------------------------------------
!*** now in my_eexpmap/dssaux_mod (shortly to be in ssclib).

!    use fitsutils_mod

    implicit none

    real(single), intent(in) :: value
    type(DSfilterT), intent(in) :: filter

    type(RangeT), pointer :: ranges(:)
    character(255)  :: gtiBlockName
    type(TableT)    :: tab
    integer         :: numGtis, i
    type(RangeT), allocatable :: gti(:)

    !-------------------------------------------------------------------
    ! Start

    valueIsPassedByDssFilter = .false.

    select case(dssFilterType(filter))
    case(RANGEFILTERT)
      ranges => dssRanges(filter)
      do i = 0, size(ranges) - 1
        if (valueIsWithinRange(value, ranges(i))) then
          valueIsPassedByDssFilter = .true.
      exit
        end if
      end do

deallocate(ranges)

!    case(REGIONFILTERT)
!      selectionExpression = selectExpression(filter)
!*** for a future version of valueIsPassedByDssFilter() with 2-value
!*** alternative input: set up pseudo-table with 1 entry, filter it,
!*** count num filtered rows.
!    case(BITMASKFILTERT)
!    case(RANGEBLOCKFILTERT)
!    case(MASKFILTERT)
!    case(REGIONBLOCKFILTERT)

    case(GTIFILTERT) ! Treat GTIs as ranges:
      gtiBlockName = dssFilterBlock(filter)
      tab = table(dataSet(gtiBlockName, READ), gtiBlockName)
      numGtis = numberOfRows(tab)
      allocate(gti(numGtis))
      call readColData(tab, 'START', gti%lower)
      call readColData(tab, 'STOP',  gti%upper)
      call release(tab)

      gti%lowerB = 1 ! = OPEN (according to dssutil.f90)
      gti%upperB = 1 ! = OPEN (according to dssutil.f90)
      do i = 1, size(gti)
        if (valueIsWithinRange(value, gti(i))) then
          valueIsPassedByDssFilter = .true.
      exit
        end if
      end do

    case default
call error('','')
    end select

  end function valueIsPassedByDssFilter

  !---------------------------------------------------------------------
  logical function valueIsWithinRange(value, range)
  !---------------------------------------------------------------------
!*** now in my_eexpmap/dssaux_mod (shortly to be in ssclib).

    implicit none

    real(single), intent(in) :: value
    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

!*** should change this so it treats OPEN and CLOSED bounds properly.
    if (range%lowerB .eq. UNDEF) then
      if (range%upperB .eq. UNDEF) then
        valueIsWithinRange = .true.
      else if (value .lt. range%upper) then
        valueIsWithinRange = .true.
      else
        valueIsWithinRange = .false.
      end if
    else
      if (range%upperB .eq. UNDEF) then
        if (value .gt. range%lower) then
          valueIsWithinRange = .true.
        else
          valueIsWithinRange = .false.
        end if
      else if (value .gt. range%lower .and. value .lt. range%upper) then
        valueIsWithinRange = .true.
      else
        valueIsWithinRange = .false.
      end if
    end if

  end function valueIsWithinRange

  !---------------------------------------------------------------------
  subroutine calcExposure(times, gti, exposureFrac)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: times(:)
    type(localGtiType), intent(in) :: gti(:)
    real(single), intent(out) :: exposureFrac(size(times))

    integer(int32) ::&
      numBins,&
      numGtis,&
      gtiNum,&
      gtiNumStart,&
      bin
    real(double) ::&
      binStartTime,&
      binStopTime,&
      currentGtiStartTime,&
      currentGtiStopTime,&
      totalGtiThisBin
    logical ::&
      firstGtiReached,&
      withinGti

    !-------------------------------------------------------------------
    ! Start

    numBins = size(times)
    numGtis = size(gti)

    ! start and stop times of first and last bins:
!*** check with TSTART and TSTOP keywords??
    binStartTime = times(1)       - 0.5 * timeBinWidth
    binStopTime  = times(numBins) + 0.5 * timeBinWidth

if (gti(1)%startTime .ge. binStopTime) call error('','')
if (gti(numGtis)%stopTime .le. binStartTime) call error('','')

    firstGtiReached = .false.
    do gtiNum = 1, numGtis
      if (gti(gtiNum)%startTime .ge. binStartTime) then
        if (gtiNum .gt. 1) then
          if (gti(gtiNum - 1)%stopTime .ge. binStartTime) then
            withinGti = .true.
            gtiNumStart = gtiNum - 1
          else
            withinGti = .false.
            gtiNumStart = gtiNum
          end if
        else
          withinGti = .false.
          gtiNumStart = gtiNum
        end if

    exit
      end if
    end do
if (.not. firstGtiReached) call error('bug0','Contact the software developer!')

    binStopTime = binStartTime

    do bin = 1, numBins
      binStartTime = binStopTime
      binStopTime  = times(bin) + 0.5 * timeBinWidth

      if (gtiNumStart .gt. numGtis) then
        exposureFrac(bin) = 0.0
    cycle
      end if

      if (withinGti) then
        currentGtiStartTime = binStartTime
      end if

      totalGtiThisBin = 0.0
      do gtiNum = gtiNumStart, numGtis
        if (.not. withinGti) then
          if (gti(gtiNum)%startTime .lt. binStopTime) then
            withinGti = .true.
            currentGtiStartTime = gti(gtiNum)%startTime
          else
            gtiNumStart = gtiNum
      exit
          end if
        end if

        if (withinGti) then
          if (gti(gtiNum)%stopTime .lt. binStopTime) then
            withinGti = .false.
            currentGtiStopTime = gti(gtiNum)%stopTime
            totalGtiThisBin = totalGtiThisBin&
            + currentGtiStopTime - currentGtiStartTime
            gtiNumStart = gtiNum + 1
          else
            gtiNumStart = gtiNum
      exit
          end if
        end if
      end do

      if (withinGti) then
        currentGtiStopTime = binStopTime
        totalGtiThisBin = totalGtiThisBin&
        + currentGtiStopTime - currentGtiStartTime
      end if

      exposureFrac(bin) = totalGtiThisBin / (binStopTime - binStartTime)
    end do ! loop over bins

  end subroutine calcExposure

  !---------------------------------------------------------------------
  subroutine dumpDSS(inSetName)
  !---------------------------------------------------------------------
!*** from colimplot

!    use dss
!    use fitsutils_mod

    implicit none

    character(len=255), intent(in) ::&
      inSetName
    type(DataSetT) ::&
      inSet
    type(BlockT) ::&
      inBlock
!    type(ArrayT) ::&
!      imageArray
    type(DssT) ::&
      my_dss
    type(DSfilterT) ::&
      filter
    type(RangeT), dimension(:), pointer :: range
    integer(kind=int16) ::&
      i
    integer(kind=int32) ::&
      bitmask
    integer :: compNum, filtNum, filterType
    logical :: typeNotFound
    character(len=20) :: dssKwdValue, columnName, filterTypeStr
    character(len=80) :: axisNames

    type filtInfo
      integer :: index
      character(len=20) :: name
    end type
    type(filtInfo) :: myFiltInfo(7)

    !-------------------------------------------------------------------
    ! Start:

    myFiltInfo(1)%index = RANGEFILTERT
    myFiltInfo(2)%index = REGIONFILTERT
    myFiltInfo(3)%index = BITMASKFILTERT
    myFiltInfo(4)%index = RANGEBLOCKFILTERT
    myFiltInfo(5)%index = GTIFILTERT
    myFiltInfo(6)%index = MASKFILTERT
    myFiltInfo(7)%index = REGIONBLOCKFILTERT
    myFiltInfo(1)%name = 'RANGEFILTERT'
    myFiltInfo(2)%name = 'REGIONFILTERT'
    myFiltInfo(3)%name = 'BITMASKFILTERT'
    myFiltInfo(4)%name = 'RANGEBLOCKFILTERT'
    myFiltInfo(5)%name = 'GTIFILTERT'
    myFiltInfo(6)%name = 'MASKFILTERT'
    myFiltInfo(7)%name = 'REGIONBLOCKFILTERT'

    inSet = dataSet(inSetName, READ)
    inBlock = block(inSet, 0, READ)
!    imageArray = array(inSet, 0, READ)
!    call getStringKwd('DSTYP1', inSet, imageArray, dssKwdValue, status)
dssKwdValue = stringAttribute(table(inSet, 'RATE'), 'DSTYP1')
!    if (status .ne. 0) then
!call error('noDSSfound','')
!    end if

    my_dss = dataSubspace(inBlock)
    do compNum = 0, dssComponents(my_dss) - 1
      print *, 'DSS component number', compNum
      do filtNum = 0, dssFilters(dssComponent(my_dss, compNum)) - 1
        filter = dssFilter(dssComponent(my_dss, compNum), filtNum)
        columnName = dssFilterName(filter)
        filterType = dssFilterType(filter)

        typeNotFound = .true.
        do i = 1, 7
          if (myFiltInfo(i)%index .eq. filterType) then
            typeNotFound = .false.
            filterTypeStr = myFiltInfo(i)%name
        exit
          end if
        end do
if (typeNotFound) call error('unknownFilterType','')

        print *, '    Filter', filtNum, columnName, filterTypeStr

        select case(filterType)
        case(RANGEFILTERT)
          range => dssRanges(filter)
          if (range(0)%lowerB .eq. UNDEF) then
            print *, '        Lower bound UNDEF'
          else
            print *, '        Lower bound =', range(0)%lower
! Note that OPEN is used both in dsslib and caloaldefs, so it can't be used.
          end if
          if (range(0)%upperB .eq. UNDEF) then
            print *, '        Upper bound UNDEF'
          else
            print *, '        Upper bound =', range(0)%upper
! Note that OPEN is used both in dsslib and caloaldefs, so it can't be used.
          end if
!        case(REGIONFILTERT)
        case(BITMASKFILTERT)
          bitmask = dssFilterMask(filter)
          print *, '        ', bitmask
!        case(RANGEBLOCKFILTERT)
!        case(GTIFILTERT)
        case(MASKFILTERT, REGIONBLOCKFILTERT)
          axisNames = dssFilterComps(filter)
          print *, '        Axis names', axisNames
        end select
      end do
      print *, ''
    end do

    call release(inSet)
  end subroutine dumpDSS

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module lccorr_lite_mod
