!
! Module: step_functions
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines used to define and manipulate a structure type StepFnT, which has 2 y values, a LH and RH one, for each x value, thus allowing functions to be defined which consist of a sequence of straight line segments with vertical steps.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module step_functions

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the '*' from the lib*.a name
! in <package>/src/Makefile.
  use types

  implicit none 

  private 

  public initializeStepY
  public addStepYVals
  public multiplyStepYVals
  public deallocateStepFnCube
  public yToStepY
  public stepIntegrate

  interface multiplyStepYVals
    module procedure multiplyStepYValsScalScal, multiplyStepYValsArrScal
  end interface

  interface stepIntegrate
    module procedure stepIntegrateCube, stepIntegrateScalar
  end interface


  type, public :: StepYT
    real(double) :: lo, hi
  end type StepYT

  type, public :: StepFnT
    real(double) :: x
    type(StepYT) :: y
  end type StepFnT

  type, public :: StepFnArrayT
    real(double)              :: x
    type(StepYT), pointer :: y(:,:)
  end type StepFnArrayT

contains 

  !---------------------------------------------------------------------
  subroutine initializeStepY(stepY)
  !---------------------------------------------------------------------

    implicit none

    type(StepYT), intent(out) :: stepY

    !-------------------------------------------------------------------
    ! Start

    stepY%lo = 0.0d0
    stepY%hi = 0.0d0

  end subroutine initializeStepY

  !---------------------------------------------------------------------
  function addStepYVals(stepYA, stepYB) result(summedStepY)
  !---------------------------------------------------------------------

    implicit none

    type(StepYT), intent(in) :: stepYA, stepYB
    type(StepYT)             :: summedStepY
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa addStepYVals'
    summedStepY%lo = stepYA%lo + stepYB%lo
    summedStepY%hi = stepYA%hi + stepYB%hi
if (testFlag) print *, '      zzz addStepYVals'

  end function addStepYVals

  !---------------------------------------------------------------------
  function multiplyStepYValsScalScal(stepYA, stepYB) result(productStepY)
  !---------------------------------------------------------------------

    implicit none

    type(StepYT), intent(in) :: stepYA, stepYB
    type(StepYT)             :: productStepY
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start
if (testFlag) print *, '      aaa multiplyStepYValsScalScal'

    productStepY%lo = stepYA%lo * stepYB%lo
    productStepY%hi = stepYA%hi * stepYB%hi
if (testFlag) print *, '      zzz multiplyStepYValsScalScal'

  end function multiplyStepYValsScalScal

  !---------------------------------------------------------------------
  function multiplyStepYValsArrScal(stepYA, stepYB) result(productStepY)
  !---------------------------------------------------------------------

    implicit none

    type(StepYT), intent(in) :: stepYA(:,:),&
                                stepYB
    type(StepYT)             :: productStepY(size(stepYA,1),size(stepYA,2))
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa multiplyStepYValsArrScal'
    productStepY%lo = stepYA%lo * stepYB%lo
    productStepY%hi = stepYA%hi * stepYB%hi
if (testFlag) print *, '      zzz multiplyStepYValsArrScal'

  end function multiplyStepYValsArrScal

  !---------------------------------------------------------------------
  subroutine deallocateStepFnCube(stepFnCube)
  !---------------------------------------------------------------------

    implicit none

    type(StepFnArrayT), pointer :: stepFnCube(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(stepFnCube)
      deallocate(stepFnCube(i)%y)
    end do
    deallocate(stepFnCube)

  end subroutine deallocateStepFnCube

  !---------------------------------------------------------------------
  function yToStepY(y) result(stepY)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: y
    type(StepYT)         :: stepY

    !-------------------------------------------------------------------
    ! Start

    stepY%lo = y
    stepY%hi = y

  end function yToStepY

  !---------------------------------------------------------------------
  function stepIntegrateCube(xSamples, inCube) result(outArray)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate outArray after use.

    implicit none

    real(double), intent(in) :: xSamples(:)
    type(StepYT), intent(in) :: inCube(:,:,:)
    real(double), pointer    :: outArray(:,:)

    integer(int16) :: numXSamples, ySize, zSize, xi

    !-------------------------------------------------------------------
    ! Start

    numXSamples = size(xSamples)
    if (numXSamples .ne. size(inCube,1))&
call error('bug23',"Please contact the code developer.")
    if (numXSamples <= 0) then
      allocate(outArray(0,0))
    end if

    ySize = size(inCube,2)
    zSize = size(inCube,3)
    allocate(outArray(ySize, zSize))
    outArray = 0.0
    do xi = 2, numXSamples
      outArray = outArray&
        + (( inCube(xi,:,:)%lo +   inCube(xi-1,:,:)%hi) / 2.0)&
        * (xSamples(xi)        - xSamples(xi-1))
    end do

  end function stepIntegrateCube

  !---------------------------------------------------------------------
  function stepIntegrateScalar(xSamples, stepY)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: xSamples(:)
    type(StepYT), intent(in) :: stepY(size(xSamples))
    real(double)             :: stepIntegrateScalar

    integer(int16) :: numSamples, xi

    !-------------------------------------------------------------------
    ! Start

    stepIntegrateScalar = 0.0 ! default
    numSamples = size(xSamples)
  if (numSamples <= 0) return

    do xi = 2, numSamples
      stepIntegrateScalar = stepIntegrateScalar&
        + ((  stepY(xi)%lo +    stepY(xi-1)%hi) / 2.0)&
        * (xSamples(xi)    - xSamples(xi-1))
    end do

  end function stepIntegrateScalar

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module step_functions
